#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'

import logging
logger = logging.getLogger('PyAMS (portal)')

import venusian
from persistent import Persistent
from persistent.mapping import PersistentMapping
from pyramid.exceptions import ConfigurationError
from pyramid.renderers import render
from pyramid.threadlocal import get_current_registry
from zope.container.contained import Contained
from zope.copy import clone
from zope.interface import alsoProvides, implementer, noLongerProvides
from zope.lifecycleevent import ObjectCreatedEvent
from zope.location import locate
from zope.schema.fieldproperty import FieldProperty
from zope.schema.vocabulary import SimpleTerm, SimpleVocabulary
from zope.traversing.interfaces import ITraversable

from pyams_cache.beaker import get_cache
from pyams_form.interfaces.form import IFormContextPermissionChecker
from pyams_portal.interfaces import IPortalContext, IPortalPage, IPortalPortletsConfiguration, IPortalTemplate, \
    IPortalTemplateConfiguration, IPortlet, IPortletConfiguration, IPortletPreviewer, IPortletRenderer, \
    IPortletRendererSettings, IPortletSettings, MANAGE_TEMPLATE_PERMISSION, PORTLET_RENDERER_SETTINGS_KEY, PREVIEW_MODE
from pyams_utils.adapter import ContextAdapter, adapter_config, get_adapter_weight, get_annotation_adapter
from pyams_utils.factory import get_object_factory, is_interface, factory_config
from pyams_utils.interfaces import ICacheKeyValue
from pyams_utils.interfaces.url import DISPLAY_CONTEXT
from pyams_utils.registry import get_global_registry
from pyams_utils.request import check_request
from pyams_utils.vocabulary import vocabulary_config
from pyams_viewlet.viewlet import ViewContentProvider


#
# Portlets utilities
#

@implementer(IPortlet)
class Portlet(object):
    """Base portlet utility"""

    permission = FieldProperty(IPortlet['permission'])

    toolbar_image = None
    toolbar_css_class = 'fa fa-fw fa-2x fa-edit'

    settings_factory = None


class portlet_config(object):
    """Class decorator used to declare a portlet"""

    venusian = venusian  # for testing injection

    def __init__(self, **settings):
        self.__dict__.update(settings)

    def __call__(self, wrapped):
        settings = self.__dict__.copy()
        depth = settings.pop('_depth', 0)

        def callback(context, name, ob):
            name = settings.get('name') or getattr(ob, 'name', None)
            if name is None:
                raise ConfigurationError("You must provide a name for a portlet")

            permission = settings.get('permission')
            if permission is not None:
                ob.permission = permission

            if type(ob) is type:
                factory = ob
                component = None
            else:
                factory = None
                component = ob

            config = context.config.with_package(info.module)
            logger.debug("Registering portlet {0} named '{1}'".format(str(component) if component else str(factory),
                                                                      name))
            config.registry.registerUtility(component=component, factory=factory,
                                            provided=IPortlet, name=name)

        info = self.venusian.attach(wrapped, callback, category='pyams_portal',
                                    depth=depth + 1)
        if info.scope == 'class':
            # if the decorator was attached to a method in a class, or
            # otherwise executed at class scope, we need to set an
            # 'attr' into the settings if one isn't already in there
            if settings.get('attr') is None:
                settings['attr'] = wrapped.__name__

        settings['_info'] = info.codeinfo  # fbo "action_method"
        return wrapped


@vocabulary_config(name='PyAMS portal portlets')
class PortletVocabulary(SimpleVocabulary):
    """Portlet vocabulary"""

    def __init__(self, context):
        request = check_request()
        translate = request.localizer.translate
        utils = request.registry.getUtilitiesFor(IPortlet)
        terms = [SimpleTerm(name, title=translate(util.label))
                 for name, util in sorted(utils, key=lambda x: translate(x[1].label))]
        super(PortletVocabulary, self).__init__(terms)


#
# Portlets renderers
#

class PortletContentProvider(ViewContentProvider):
    """Base portlet content provider"""

    def __init__(self, context, request, view, settings):
        super(PortletContentProvider, self).__init__(context, request, view)
        self.portlet = request.registry.queryUtility(IPortlet, name=settings.configuration.portlet_name)
        self.settings = settings

    def render(self):
        if self.portlet is None:
            return ''
        if self.portlet.permission and not self.request.has_permission(self.portlet.permission):
            return ''
        return super(PortletContentProvider, self).render()


@implementer(IPortletPreviewer)
class PortletPreviewer(PortletContentProvider):
    """Portlet previewer adapter"""

    @property
    def slot_configuration(self):
        if IPortalTemplate.providedBy(self.context):
            template = self.context
        else:
            template = IPortalPage(self.context).template
        config = IPortalTemplateConfiguration(template)
        slot_id, slot_name = config.get_portlet_slot(self.settings.configuration.portlet_id)
        return config.get_slot_configuration(slot_name)

    def render(self):
        if self.settings.renderer == 'hidden':
            return render('templates/portlet-hidden.pt', {}, request=self.request)
        else:
            return super(PortletPreviewer, self).render()


PORTLETS_CACHE_REGION = 'portlets'
PORTLETS_CACHE_NAME = 'PyAMS::portlet'

PORTLETS_CACHE_KEY = 'portlet::{hostname}::{portlet}::{context}::{lang}'
PORTLETS_CACHE_DISPLAY_CONTEXT_KEY = 'portlet::{hostname}::{portlet}::{context}::{display}::{lang}'


@implementer(IPortletRenderer)
class PortletRenderer(PortletContentProvider):
    """Portlet renderer adapter"""

    settings_interface = None

    @property
    def settings_key(self):
        return PORTLET_RENDERER_SETTINGS_KEY.format(self.settings.renderer)

    target_interface = None
    use_authentication = False

    weight = 0

    @property
    def slot_configuration(self):
        template = IPortalPage(self.context).template
        config = IPortalTemplateConfiguration(template)
        slot_id, slot_name = config.get_portlet_slot(self.settings.configuration.portlet_id)
        return config.get_slot_configuration(slot_name)

    @property
    def renderer_settings(self):
        return IPortletRendererSettings(self.settings)

    @property
    def use_portlets_cache(self):
        return not bool(self.request.params)

    def render(self):
        preview_mode = self.request.annotations.get(PREVIEW_MODE, False)
        if preview_mode or not self.use_portlets_cache:
            return super(PortletRenderer, self).render()
        else:
            portlets_cache = get_cache(PORTLETS_CACHE_REGION, PORTLETS_CACHE_NAME)
            # check for display context
            display_context = self.request.annotations.get(DISPLAY_CONTEXT)
            if display_context is None:
                cache_key = PORTLETS_CACHE_KEY.format(portlet=ICacheKeyValue(self.settings),
                                                      hostname=self.request.host,
                                                      context=ICacheKeyValue(self.context),
                                                      lang=self.request.locale_name)
            else:
                cache_key = PORTLETS_CACHE_DISPLAY_CONTEXT_KEY.format(portlet=ICacheKeyValue(self.settings),
                                                                      hostname=self.request.host,
                                                                      context=ICacheKeyValue(self.context),
                                                                      display=ICacheKeyValue(display_context),
                                                                      lang=self.request.locale_name)
            if self.use_authentication:
                cache_key = '{0}::{1}'.format(cache_key, self.request.principal.id)
            # load rendered content from cache, or create output and store it in cache
            try:
                result = portlets_cache.get_value(cache_key)
                logger.debug("Retrieving portlet content from cache key {0}".format(cache_key))
            except KeyError:
                self.update()
                result = super(PortletRenderer, self).render()
                portlets_cache.set_value(cache_key, result)
                logger.debug("Storing portlet content to cache key {0}".format(cache_key))
            return result


@vocabulary_config(name='PyAMS portlet renderers')
class PortletRenderersVocabulary(SimpleVocabulary):
    """Portlet renderers vocabulary"""

    def __init__(self, context):
        request = check_request()
        translate = request.localizer.translate
        terms = [SimpleTerm(name, title=translate(adapter.label))
                 for name, adapter in sorted(request.registry.getAdapters((request.root, request, request, context),
                                                                          IPortletRenderer),
                                             key=get_adapter_weight)]
        super(PortletRenderersVocabulary, self).__init__(terms)


#
# Portlet configuration
#

@implementer(IPortletSettings)
class PortletSettings(Persistent, Contained):
    """Portlet settings persistent class

    This class is supposed to be sub-classed by all custom portlet subclasses to
    store their configuration settings.

    Each portlet sub-class must define it's settings factory in it's "settings_factory" attribute.
    Given factory can be a function, a class or an interface; in this last case, implementation
    is looking for default object factory registered for this interface.
    """

    _renderer = FieldProperty(IPortletSettings['renderer'])

    __name__ = '++settings++'

    def __init__(self, configuration=None):
        self.__parent__ = configuration

    @property
    def visible(self):
        return self._renderer != 'hidden'

    @property
    def renderer(self):
        return self._renderer

    @renderer.setter
    def renderer(self, value):
        value = value or ''
        if value == self._renderer:
            return
        request = check_request()
        renderer = request.registry.queryMultiAdapter((request.root, request, request, self), IPortletRenderer,
                                                      name=self._renderer)
        if (renderer is not None) and (renderer.target_interface is not None):
            noLongerProvides(self, renderer.target_interface)
        self._renderer = value
        renderer = request.registry.queryMultiAdapter((request.root, request, request, self), IPortletRenderer,
                                                      name=self._renderer)
        if (renderer is not None) and (renderer.target_interface is not None):
            alsoProvides(self, renderer.target_interface)

    def get_renderer(self, request=None):
        if request is None:
            request = check_request()
        return request.registry.queryMultiAdapter((request.root, request, request, self), IPortletRenderer,
                                                  name=self._renderer)

    @property
    def configuration(self):
        return self.__parent__

    @configuration.setter
    def configuration(self, value):
        if self.__parent__ is None:
            self.__parent__ = value


@adapter_config(context=IPortletSettings, provides=IFormContextPermissionChecker)
class PortletSettingsPermissionChecker(ContextAdapter):
    """Portlet settings permission checker"""

    edit_permission = MANAGE_TEMPLATE_PERMISSION


@adapter_config(context=IPortletSettings, provides=IPortletRendererSettings)
def portlet_renderer_settings_adapter(context):
    renderer = context.get_renderer()
    if not renderer.settings_interface:
        return None
    return get_annotation_adapter(context, renderer.settings_key, renderer.settings_interface,
                                  name='++renderer++')


@adapter_config(name='renderer', context=IPortletSettings, provides=ITraversable)
class PortletSettingsRendererSettingsTraverser(ContextAdapter):
    """Portlet settings traverser to renderer settings"""

    def traverse(self, name, furtherpath=None):
        return IPortletRendererSettings(self.context)


@implementer(IPortletConfiguration)
class PortletConfiguration(Persistent, Contained):
    """Portlet configuration persistent class

    This class is a generic persistent class which is used to store all portlet
    configuration and is *not* supposed to be sub-classed.

    PortletConfiguration.__parent__ points to context where configuration is applied (each context or
    local template).
    PortletConfiguration.parent points to context from where configuration is inherited.
    """

    portlet_id = FieldProperty(IPortletConfiguration['portlet_id'])
    portlet_name = None
    _inherit_parent = FieldProperty(IPortletConfiguration['inherit_parent'])
    _settings = FieldProperty(IPortletConfiguration['settings'])

    def __init__(self, portlet):
        self.portlet_name = portlet.name
        factory = portlet.settings_factory
        if is_interface(factory):
            factory = get_object_factory(factory)
        assert factory is not None, "Missing portlet settings factory"
        settings = factory()
        settings.configuration = self
        self._settings = settings

    def get_portlet(self):
        return get_global_registry().queryUtility(IPortlet, name=self.portlet_name)

    @property
    def can_inherit(self):
        return not IPortalTemplate.providedBy(self.__parent__)

    @property
    def inherit_parent(self):
        return self._inherit_parent if self.can_inherit else False

    @inherit_parent.setter
    def inherit_parent(self, value):
        if (not value) or self.can_inherit:
            self._inherit_parent = value

    @property
    def parent(self):
        parent = self.__parent__
        if IPortalTemplate.providedBy(parent):
            return parent
        while True:
            if IPortalContext.providedBy(parent):
                configuration = IPortalPortletsConfiguration(parent).get_portlet_configuration(self.portlet_id)
                if not configuration.inherit_parent:
                    return parent
                page = IPortalPage(parent)
                if not page.inherit_parent:
                    break
            parent = parent.__parent__
            if parent is None:
                break
        page = IPortalPage(parent, None)
        if page is not None:
            return page.template

    @property
    def settings(self):
        if self.inherit_parent:
            return IPortalPortletsConfiguration(self.parent).get_portlet_configuration(self.portlet_id).settings
        else:
            return self._settings

    @property
    def editor_settings(self):
        return self._settings

    def get_settings(self, allow_inherit=True):
        if allow_inherit:
            return self.settings
        else:
            return self._settings


@adapter_config(context=IPortlet, provides=IPortletConfiguration)
def portlet_configuration_adapter(portlet):
    """Portlet configuration factory"""
    return PortletConfiguration(portlet)


@adapter_config(context=IPortletConfiguration, provides=IPortletSettings)
def portlet_configuration_settings_adapter(configuration):
    """Portlet configuration settings adapter"""
    return configuration.settings


@adapter_config(name='settings', context=IPortletConfiguration, provides=ITraversable)
class PortletConfigurationSettingsTraverser(ContextAdapter):
    """++settings++ portlet configuration traverser"""

    def traverse(self, name, furtherpath=None):
        return self.context.settings


@adapter_config(context=IPortletConfiguration, provides=IFormContextPermissionChecker)
class PortletConfigurationPermissionChecker(ContextAdapter):
    """Portlet configuration permission checker"""

    edit_permission = MANAGE_TEMPLATE_PERMISSION


#
# Template portlets configuration
#

@factory_config(IPortalPortletsConfiguration)
class PortalPortletsConfiguration(PersistentMapping, Contained):
    """Portal portlets configuration"""

    @classmethod
    def clone(cls, source_config, new_parent):
        """Clone source configuration"""
        configuration = source_config.__class__()
        get_current_registry().notify(ObjectCreatedEvent(configuration))
        locate(configuration, new_parent)
        for config_id, config_portlet in source_config.items():
            config = clone(config_portlet)
            configuration[config_id] = config
        return configuration

    def __setitem__(self, key, value):
        super(PortalPortletsConfiguration, self).__setitem__(key, value)
        locate(value, self.__parent__, '++portlet++{0}'.format(key))

    def get_portlet_configuration(self, portlet_id):
        configuration = self.get(portlet_id)
        if configuration is None:
            if IPortalTemplate.providedBy(self.__parent__):
                portlets = IPortalPortletsConfiguration(self.__parent__)
            else:
                template = IPortalPage(self.__parent__).template
                portlets = IPortalPortletsConfiguration(template)
            configuration = clone(portlets.get_portlet_configuration(portlet_id))
            get_current_registry().notify(ObjectCreatedEvent(configuration))
            self.set_portlet_configuration(portlet_id, configuration)
        return configuration

    def set_portlet_configuration(self, portlet_id, config):
        config.portlet_id = portlet_id
        self[portlet_id] = config

    def delete_portlet_configuration(self, portlet_id):
        if isinstance(portlet_id, int):
            portlet_id = (portlet_id,)
        for p_id in portlet_id:
            del self[p_id]
